#include "config.h"

#ifdef HTTP_SERVER_ENABLE

#include "http_server.h"
#include "effects.h"
#include "my_wifi.h"
#include "functions.h"
#include "prototypes.h"
#include <FS.h>

AsyncWebServer http_server(HTTP_SERVER_PORT);
AsyncWebSocket ws("/ws");

File upload_file;

/*void http_server_handle_file_upload() {
	if (http_server.uri() != "/upload") return;

	HTTPUpload upload = http_server.upload();

	if (upload.status == UPLOAD_FILE_START) {
		String filename = upload.filename;
		if (!filename.startsWith("/")) filename = "/" + filename;
		LOGln("HTTP * Upload of %s starting...", upload.filename.c_str());
		upload_file = SPIFFS.open(filename, "w");
	} else if (upload.status == UPLOAD_FILE_WRITE) {
		if (upload_file) upload_file.write(upload.buf, upload.currentSize);
	} else if (upload.status == UPLOAD_FILE_END) {
		if (upload_file) upload_file.close();
		LOGln("HTTP * Upload of %s with %d bytes done.", upload.filename.c_str(), upload.totalSize);
	}
}*/

void handle_ws(AsyncWebSocket* server, AsyncWebSocketClient* client, AwsEventType type, void* arg, uint8_t* data, size_t len) {
	if (type == WS_EVT_CONNECT) {
		LOGln("Websocket * Client connected.");
	} else if (type == WS_EVT_DISCONNECT) {
		LOGln("Websocket * Client disconnected.");
	}
}

void http_server_setup() {
	PGM_P text_plain = PSTR("text/plain");
	PGM_P text_html = PSTR("text/html");
	http_server.on("/", HTTP_GET, [&](AsyncWebServerRequest* request){
		LOGln("HTTP * GET /");
		String message = "<html><head><title>Pitrix</title></head><body><h1>Pitrix</h1><p><a href='/settings'>Settings</a></p><p><a href='/effects'>Effect</a></p><p>Known animations:</p>";
		if (!SPIFFS.begin()) {
			message += "<strong>No SPIFFS file system found.</strong>";
		} else {
			message += "<ul>";
			Dir dir = SPIFFS.openDir("/");
			while (dir.next()) {
				message += "<li>" + dir.fileName() + " (<a href='/delete?" + dir.fileName() + "'>delete</a>)</li>";
			}
			message += "</ul>";
			message += "<form action='/upload' method='POST'><input type='file' name='file' /><input type='submit' value='Upload file' /></form>";
		}
		message += "</body></html>";
		request->send(200, "text/html", message);
	});
	http_server.on("/settings", HTTP_GET, [&](AsyncWebServerRequest* request) {
		String message = "<html><head><title>Pitrix settings</title></head><body><h1>Pitrix settings</h1><a href='/'>Back to main page</a><table>\n";
		for (int i=0; i<all_settings_size; i++) {
			Setting s = all_settings[i];
			uint16_t default_value = setting_default(&s);
			uint16_t value = *(s.value);

			message += "<tr><td>";
			if (default_value != value) {
				message += "<strong>";
			}
			message += s.name;
			if (default_value != value) {
				message += "<strong>";
			}
			message += "</td><td>";
			message += value;
			if (default_value != value) {
				message += " (";
				message += default_value;
				message += ")";
			}
			message += "</td><td><form method='POST' action='/settings?redir=1'><input type='hidden' name='key' value='";
			message += s.name;
			message += "'/>";
			if (s.type==TYPE_UINT8 || s.type==TYPE_UINT16) {
				message += "<input type='number' name='value' value='";
				message += value;
				message += "' min='0' max='";
				if (s.type==TYPE_UINT8) {
					message += "255";
				} else {
					message += "65535";
				}
				message += "' />";
			} else if (s.type==TYPE_BOOL) {
				message += "<input type='radio' name='value' value='0'";
				if (value==0) {
					message += " checked='checked'";
				}
				message += "> Off / <input type='radio' name='value' value='1'";
				if (value==1) {
					message += " checked='checked'";
				}
				message += "> On";
			}
			message += "<input type='submit' value='Save' />";
			message += "</form></td></tr>\n";
		}
		message += "</table></body></html>";
		request->send(200, "text/html", message);
	});
	http_server.on("/settings", HTTP_POST, [&](AsyncWebServerRequest* request) {
		if (!request->hasParam("key", true) || !request->hasParam("value", true)) {
			request->send(400, "text/plain", "Missing argument.");
			return;
		}
		String name = request->getParam("key", true)->value();
		uint16_t value = request->getParam("value", true)->value().toInt();

		if (change_setting(name.c_str(), value)) {
			if (request->hasParam("redir")) {
				AsyncWebServerResponse* response = request->beginResponse(301, "text/plain", "Moved");
				response->addHeader("Location", "/settings");
				request->send(response);
			} else {
				request->send(200, "text/plain", "OK");
			}
			save_settings();
		} else {
			request->send(400, "text/plain", "Could not change setting.");
		}
	});
	http_server.on("/settings/load", HTTP_POST, [&](AsyncWebServerRequest* request) {
		load_settings();
		request->send(200, "text/plain", "OK");
	});
	http_server.on("/settings/save", HTTP_POST, [&](AsyncWebServerRequest* request) {
		save_settings();
		request->send(200, "text/plain", "OK");
	});
	http_server.on("/settings.txt", HTTP_GET, [&](AsyncWebServerRequest* request) {
		File f;
		if (SPIFFS.begin()) {
			f=SPIFFS.open("/pitrix_settings.txt", "r");
			if (f) {
				String s = f.readString();
				f.close();
				request->send(200, "text/plain", s);
				return;
			}
		}
		request->send(500, "text/plain", "Could not read settings.");
	});
	http_server.on("/effects", HTTP_GET, [&](AsyncWebServerRequest* request) {
		String message = "<html><head><title>Pitrix effects</title></head><body><h1>Pitrix settings</h1><a href='/'>Back to main page</a><table>";
		for (int i=0; i<effects_size; i++) {
			message += "<tr><td>";
			message += effects[i].name;
			message += "</td><td><form method='post' action='/effects'><input type='hidden' name='name' value='";
			message += effects[i].name;
			message += "'><input type='hidden' name='redir' value='1'><input type='submit' value='Select'></form></td></tr>";
		}
		message += "</table></body></html>";
		request->send(200, "text/html", message);
	});
	http_server.on("/effects", HTTP_POST, [&](AsyncWebServerRequest* request) {
		if (!request->hasParam("name", true)) {
			request->send(400, "text/plain", "Missing argument 'name'.");
			return;
		}
		String name = request->getParam("name", true)->value();
		if (change_current_effect(name)) {
			if (request->hasParam("redir")) {
				AsyncWebServerResponse* response = request->beginResponse(301, "text/plain", "Moved");
				response->addHeader("Location", "/effects");
				request->send(response);
			} else {
				request->send(200, "text/plain", "OK");
			}
		} else {
			request->send(404, "text/plain", "Effect not found.");
		}
	});
	http_server.on("/delete", HTTP_GET, [&](AsyncWebServerRequest* request) {
		LOGln("HTTP * GET /delete");
		if (request->params()==0) {
			request->send_P(400, text_plain, PSTR("No filename given"));
			return;
		}
		String file = request->getParam(0)->value();
		if (file == "/") {
			request->send_P(400, text_plain, PSTR("Invalid path"));
			return;
		}
		if (!SPIFFS.exists(file)) {
			request->send_P(400, text_plain, PSTR("File does not exist."));
			return;
		}
		SPIFFS.remove(file);
		request->send_P(200, text_plain, PSTR("OK"));
	});
	/*http_server.on("/upload", HTTP_POST, [](AsyncWebServerRequest* request) {
		LOGln("HTTP * POST /upload");
		request->send(200, "text/plain", "OK");
	}, http_server_handle_file_upload);*/
	http_server.on("/free_heap", HTTP_GET, [&](AsyncWebServerRequest* request){
		LOGln("HTTP * GET /free_heap");
		request->send(200, "text/plain", String(ESP.getFreeHeap()));
	});
	http_server.on("/uptime", HTTP_GET, [&](AsyncWebServerRequest* request){
		LOGln("HTTP * GET /uptime");
		request->send(200, "text/plain", String(millis()/1000));
	});
	http_server.on("/fps", HTTP_GET, [](AsyncWebServerRequest* request){
		LOGln("HTTP * GET /fps");
		request->send(200, "text/plain", String(FastLED.getFPS()));
	});
	http_server.on("/reboot", HTTP_POST, [](AsyncWebServerRequest* request){
		LOGln("HTTP * POST /reboot");
		ESP.restart();
	});
	http_server.on("/mode", HTTP_POST, [&](AsyncWebServerRequest* request){
		LOGln("HTTP * POST /mode with value %s", request->getParam("plain", true)->value().c_str());
		if (!request->hasParam("plain", true)) {
			request->send_P(400, text_plain, PSTR("No effect given."));
			return;
		}
		String val = request->getParam("plain", true)->value();
		if (change_current_effect(val)) {
			request->send(200, "text/plain", "OK");
		} else {
			request->send_P(400, text_plain, PSTR("Unknown effect."));
		}
	});
	http_server.on("/monitor", HTTP_GET, [&](AsyncWebServerRequest* request) {
		static const char* html PROGMEM = R"(
		<html>
			<head>
				<title>Pitrix</title>
			</head>
			<body>
				<canvas id='target' width='800' height='500'></canvas>
				<script type='text/javascript'>
					ctx = document.getElementById('target').getContext('2d');
					socket = new WebSocket((document.location.protocol=='https:' ? 'wss' : 'ws') + '://' + document.location.host + '/ws');
					socket.binaryType = 'arraybuffer';
					socket.onmessage = function(message) {
						var buffer = new Uint8Array(message.data);
						var width = buffer[0];
						var height = buffer[1];
						if (buffer[2] != 255) return;
						var offset = 3;
						ctx.fillStyle = '#000000';
						ctx.fillRect(0, 0, 20*width, 20*height);
						for (var y=0; y<height; y++) for (var x=0; x<width; x++) {
							var r = buffer[offset + 0];
							var g = buffer[offset + 1];
							var b = buffer[offset + 2];
							offset = offset + 3;
							ctx.beginPath();
							ctx.arc(x*20 + 10, y*20 + 10, 10, 0, 2*Math.PI, false);
							ctx.fillStyle = 'rgb('+r+','+g+','+b+')';
							ctx.fill();
						}
					};
					function redraw() {
						ctx.clearRect(0, 0, ctx.canvas.width, ctx.canvas.height);
					}
				</script>
			</body>
		</html>
		)";
		
		request->send_P(200, PSTR("text/html"), html);
	});
	
	
	ws.onEvent(handle_ws);
	http_server.addHandler(&ws);
	
	
	http_server.begin();

	MDNS.addService("_http", "_tcp", 80);
}

void http_server_send_framedata() {
	if (ws.count() > 0) {
		uint16_t _size = LED_WIDTH * LED_HEIGHT * 3 + 4;
		uint8_t* _buffer = new uint8_t[_size];
		_buffer[0] = LED_WIDTH;
		_buffer[1] = LED_HEIGHT;
		_buffer[2] = 255;
		for (uint8_t y=0; y<LED_HEIGHT; y++) for(uint8_t x=0; x<LED_WIDTH; x++) {
			uint16_t index = XYsafe(x, y);
			CRGB pixel = leds[index];
			_buffer[3 + (y*LED_WIDTH + x)*3 + 0] = (pixel.r==255 ? 254 : pixel.r);
			_buffer[3 + (y*LED_WIDTH + x)*3 + 1] = (pixel.g==255 ? 254 : pixel.g);
			_buffer[3 + (y*LED_WIDTH + x)*3 + 2] = (pixel.b==255 ? 254 : pixel.b);
		}
		_buffer[_size - 1] = 255;
		ws.binaryAll(_buffer, _size);
		delete _buffer;
	}
}

#endif
/*
#include <ESP8266WiFi.h>
#include <WebSocketsServer.h>
#include "config.h"
#include "websockets.h"
#include "my_fastled.h"
#include "functions.h"
#include "effects.h"

WebSocketsServer ws = WebSocketsServer(81);

void ws_event(uint8_t num, WStype_t type, uint8_t* payload, size_t length) {
	if (type == WStype_CONNECTED) {
		LOGln("Websockets * Client connected.");
	} else if (type == WStype_DISCONNECTED) {
		LOGln("Websockets * Client disconnected.");
	} else if (type == WStype_TEXT) {
		String msg = String((char*)payload);
		LOGln("Websockets * Received: %s", msg.c_str());
		
		if (msg.startsWith("E:")) {
			change_current_effect(msg.substring(2));
		}
	}
}

void websocket_setup() {
	ws.begin();
	ws.onEvent(ws_event);
}

void send_data() {
	uint16_t _size = LED_WIDTH * LED_HEIGHT * 3 + 4;
	uint8_t* _buffer = new uint8_t[_size];
	_buffer[0] = LED_WIDTH;
	_buffer[1] = LED_HEIGHT;
	_buffer[2] = 255;
	for (uint8_t y=0; y<LED_HEIGHT; y++) for(uint8_t x=0; x<LED_WIDTH; x++) {
		uint16_t index = XYsafe(x, y);
		CRGB pixel = leds[index];
		_buffer[3 + (y*LED_WIDTH + x)*3 + 0] = (pixel.r==255 ? 254 : pixel.r);
		_buffer[3 + (y*LED_WIDTH + x)*3 + 1] = (pixel.g==255 ? 254 : pixel.g);
		_buffer[3 + (y*LED_WIDTH + x)*3 + 2] = (pixel.b==255 ? 254 : pixel.b);
	}
	_buffer[_size - 1] = 255;
	ws.broadcastBIN(_buffer, _size);
	delete _buffer;
}

void websocket_loop() {
	if (ws.connectedClients()>0) {
		send_data();
	}
	ws.loop();
}
*/
